;;;; Tasks are scheuled with queues.  

(declare (usual-integrations))

;;; There are queues of tasks for individual actors, and 
;;; there is a queue for runnable actors (those who have 
;;; non-empty task queues).  A task is a thunk.

(define runnable-actors)
(define (init-actors)
  (set! runnable-actors (queue:make)))

(define (switch-threads working-continuation)
  (define (get-task)
    (if (queue:empty? runnable-actors)
        'nothing-to-do
        (let ((actor (queue:get-first! runnable-actors)))
	  (if (actor-procedure? actor)
              (let ((actor-task-queue (get-actor-task-queue actor)))
                ;; If there are no more tasks for this actor,
                ;; it is not runnable.
                (if (queue:empty? actor-task-queue) ; Nothing for actor.
                    (begin (set-actor-runnable! actor #f)
                           'try-again)
                    (let ((task (queue:get-first! actor-task-queue)))
		      ;; Got the first task, and took it off the queue.
		      ;; If there are no more tasks for this actor,
		      ;; it is not runnable.
		      ;; If there are more tasks for this actor,
		      ;; requeue it.
                      (if (queue:empty? actor-task-queue)
			  (set-actor-runnable! actor #f)
			  (queue:add-to-end! runnable-actors actor))
		      ;; Return task to be executed.
		      task)))
              (lambda () (actor 'go))))))
  (let lp ((task (atomically get-task)))
    (cond ((eq? task 'nothing-to-do)
	   (working-continuation 'resume))
	  ((eq? task 'try-again)
	   (lp (atomically get-task)))
	  (else
	   (task)
	   (working-continuation 'resume)))))

;;; To add a task to an actor we must add it to the actor's task
;;; queue and we must make the actor runnable.  This must all be
;;; done atomically.

(define (add-to-tasks! actor task)
  (atomically
    (lambda ()
      (queue:add-to-end! (get-actor-task-queue actor) task)
      (add-to-runnable! actor)))
  'task-added)

(define (add-to-runnable! actor)
  (if (actor-runnable? actor)
      'already-runnable
      (begin (set-actor-runnable! actor #t)
	     (queue:add-to-end! runnable-actors actor)
	     'made-runnable)))

;;; We use the queue design similar to SICP Section 3.3.2
;;; Although the operations that are done here must be atomic, 
;;; the atomicity is imposed at the calls to these routines.

(define-record-type queue
    (queue:make-record front-ptr rear-ptr)
    queue?
  (front-ptr queue:front-ptr queue:set-front-ptr!)
  (rear-ptr  queue:rear-ptr  queue:set-rear-ptr!))

(define (queue:make)
  (queue:make-record '() '()))

(define (queue:empty? queue)
  (null? (queue:front-ptr queue)))

(define (queue:get-first! queue)
  (if (null? (queue:front-ptr queue))
      (error "get-first! called with an empty queue" queue)
      (let ((first (car (queue:front-ptr queue)))
	    (rest (cdr (queue:front-ptr queue))))
	(queue:set-front-ptr! queue rest)
	(if (null? rest)
	    (queue:set-rear-ptr! queue '()))
	first)))

(define (queue:add-to-end! queue item)
  (let ((new-pair (cons item '())))
    (cond ((null? (queue:front-ptr queue))
	   (queue:set-front-ptr! queue new-pair)
	   (queue:set-rear-ptr! queue new-pair))
	  (else
	   (set-cdr! (queue:rear-ptr queue) new-pair)
	   (queue:set-rear-ptr! queue new-pair))))
  'done)

